/*
** Module   : BrowseColumnWidget.java
** Abstract : server side browse column widget implementation
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20051026   @23480 Created initial version.
** 002 SIY 20060227   @25122 Changes for the browse controller 
**                           implementation.
** 003 SIY 20060330   @25278 Added methods for LABEL attribute.
** 004 SIY 20060417   @26218 Added setColumnLabel() method.
** 005 GES 20060803   @28365 Major cleanup and refactoring.
** 006 GES 20080320   @37580 Added TYPE attribute.
** 007 SIY 20080603   @38474 setColumn() now operates on separate
**                           COLUMN-LABEL attribute. 
** 008 SIY 20090206   @41262 Changed setLabel() logic to make it properly
**                           handling setLabel/setColumnLabel/setLavel 
**                           sequences (this can happen when label is
**                           dynamically set after browse construction).
** 009 SIY 20090218   @41328 Added support for AT base field format width.
** 010 SIY 20111101          Updated imports.
** 011 CA  20130221          Refactored usage of PRIVATE-DATA and TYPE attribute.
** 012 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 013 CA  20130927          Added LegacyResource annotation. TODO: implement the resource type 
**                           based actual column type and "-browcoltype" parameter.
** 014 CA  20131115          Changes related to properly instantiate dynamic resources.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * Server side representation of a browse column.
 * <p>
 * The {@link LegacyResource} annotation targets the {@link LegacyResource#FILL_IN} type.
 * <p>
 * The 4GL documentation states that {@link LegacyResource#BROWSE_COLUMN} type is used only when 
 * "-browcoltype" parameter is * at program startup.  Else, the type can be either 
 * {@link LegacyResource#FILL_IN}, {@link LegacyResource#COMBO_BOX} or 
 * {@link LegacyResource#TOGGLE_BOX} (depending on the column's widget).
 * 
 * TODO: fix this to report the correct type. 
 */
@LegacyResource(resource = LegacyResource.BROWSE_COLUMN)
public class BrowseColumnWidget
extends GenericWidget
{
   /** Back reference to the variable accessor. */
   private FrameElement element = null;
   
   /** Reference to our configuration which is pre-cast to the proper type. */
   private BrowseColumnConfig bcc = null;

   /**
    * Default constructor
    */
   public BrowseColumnWidget()
   {
      super(false, new BrowseColumnConfig(ComponentConfig.BROWSE_COLUMN, null));
      bcc = (BrowseColumnConfig) config;
   }
   
   /**
    * Set the COLUMN-LABEL attribute.
    * 
    * @param    label
    *           The new value for the attribute.
    */
   public void setColumnLabel(String label)
   {
      bcc.setColumnLabel(label);
      pushScreenDefinition();
   }
   
   /**
    * Sets the label text, removing any '&' characters first (this is
    * an undocumented feature of Progress with labels).
    * 
    * @param    label
    *           The new label text.
    */
   public void setLabel(String label)
   {
      if (bcc.getLabel() != null)
         bcc.setColumnLabel(label);
      
      super.setLabel(label);
   }

   /**
    * Set the DATA-TYPE attribute value.
    * 
    * @param    value
    *           New value for the attribute.
    */
   public void setDataType(String value)
   {
      bcc.setDataType(value);
      pushScreenDefinition();
   }

   /**
    * Set the WIDTH attribute value.
    * 
    * @param    value
    *           New value for the attribute.
    */
   public void setWidth(int value)
   {
      bcc.setAutoResize(false);
      bcc.setWidthChars(value);
      pushScreenDefinition();
   }

   /**
    * Set format width of AT base field.
    * 
    * @param   value
    *          New format width.
    */
   public void setFormatWidth(int value)
   {
      bcc.setFormatWidth(value);
      // We do not push definition here because
      // this is not a run-time changeable value, 
      // it set only once during startup.
   }
   
   /**
    * Enable editing of the column.
    * 
    * @param    value
    *           New value for the attribute.
    */
   public void setEditable(boolean value)
   {
      bcc.setReadOnly(!value);
      pushScreenDefinition();
   }

   /**
    * Set AUTO-ZAP attribute.
    * 
    * @param    value
    *           New value for the attribute.
    */
   public void setAutoZap(boolean value)
   {
      bcc.setAutoZap(value);
      pushScreenDefinition();
   }

   /**
    * Set LABELS attribute.
    * 
    * @param    labels
    *           New value for the attribute.
    */
   public void setLabels(boolean labels)
   {
      bcc.setLabels(labels);
      pushScreenDefinition();
   }

   /**
    * Set LABELS attribute as opposite to passed value.
    * 
    * @param    nolabels
    *           New value for the attribute.
    */
   public void setNoLabels(boolean nolabels)
   {
      setLabels(!nolabels);
   }
   
   /**
    * Force presence of the label.
    * 
    * @param    force
    *           New value for the flag.
    */
   public void setForceLabel(boolean force)
   {
      bcc.setForceLabel(force);
      pushScreenDefinition();
   }
   
   /**
    * Provide link to the backing accessor.
    * 
    * @param    element
    *           The accessor for this column.
    */
   protected void setElement(FrameElement element)
   {
      this.element = element;
   }
   
   /**
    * Retrieve value of the associated variable.
    * 
    * @return   Current value of the variable.
    */
   public BaseDataType getValue()
   {
      return element.get();
   }

   /**
    * Assign value to associated variable.
    * 
    * @param    value
    *           New value for the variable.
    */
   public void setValue(BaseDataType value)
   {
      element.set(value);
   }

   /**
    * Set ID of the browse widget to which this column belongs.
    * 
    * @param    id
    *           Browse ID.
    */
   public void setBrowseId(int id)
   {
      bcc.setBrowseId(id);
   }
}
