/*
** Module   : AccumulatorManager.java
** Abstract : provides block-behavior for the accumulators.
**
** Copyright (c) 2008-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 CA  20080129   @37024 Created initial version. Provides context
**                           local scoped accumulators.
** 002 CA  20080214   @37130 The set for each block must contain all
**                           accumulator instances which can be accesed
**                           from that block. Removed master set. Now,
**                           when a new scope is started, the set for the
**                           new scope is added all known accumulators
**                           from enclosing scope.
** 003 GES 20080625   @38956 Moved to new BlockType enum.
** 004 ECF 20080630   @39025 Fixed memory leak. Some Accumulators were
**                           being added to a block scope one level above
**                           where they were first being used. If that
**                           scope was never popped, but the lower scope
**                           was, these instances would be dead, but still
**                           doing work during scopeStart(), creating a
**                           CPU drain.
** 005 ECF 20080715   @39126 Fixed NPE. Created default WorkArea ctor
**                           which initializes deque and pushes the first
**                           entry.
** 006 GES 20080917   @39849 Minor class name change.
** 007 CA  20131013          Added no-op scopeDeleted() method, required by the changes in 
**                           Scopeable interface.
*/

package com.goldencode.p2j.util;

import java.util.*;
import com.goldencode.p2j.security.ContextLocal;

/**
 * Provides context-local scoped accumulators.
 * <p>
 * All references can be added to the top scope (closest enclosing scope).  
 * <p>
 * Implements the <code>Scopeable</code> interface to allow this class to be
 * automatically notified of all scope start and end occurances. This removes
 * the need for client code to be instrumented to provide this service. The
 * use of this class is now tied to the {@link TransactionManager} class. In
 * particular, the singleton instance of this class is used in each context-
 * local list of <code>Scopeable</code> objects as a dispatcher. This bit
 * of nastiness is necessary since the Java Interface can't be implemented
 * on a static basis but instead requires an instance method implementation.
 * These instance methods then call the static counterparts which is where
 * the content-local versions of the resource pools are accessed.
 * <p>
 * The initialization processing is implemented in the {@link #initialize}
 * method and it is called during server startup based on an entry in the
 * directory until the path "/server/&lt;serverId&gt;/startup" or in
 * "/server/default/startup".
 *
 * @author    CA
 */
public class AccumulatorManager
implements Scopeable,
           ScopeableFactory
{
   /** Pool of accums which is context-local. Initializes upon first use. */ 
   private static ContextContainer context = new ContextContainer();
   
   /** Singleton instance of this class (across entire JVM, all contexts). */ 
   private static AccumulatorManager single = null;
   
   /**
    * Implements the singleton nature of this class, constructs this instance
    * if it hasn't already been constructed.
    *
    * @return   The single instance of this class.
    */
   public static synchronized AccumulatorManager getInstance()
   {
      if (single == null)
      {
         single = new AccumulatorManager();
      }
      
      return single;
   }
   
   /**
    * Initialize the accumulator framework. This registers with the
    * {@link TransactionManager TransactionManager} a factory object which
    * creates a <code>Scopeable</code> instance.
    *
    * @throws  IllegalStateException
    *          If invoked more than once.
    */
   public static synchronized void initialize()
   {
      if (single == null)
      {
         TransactionManager.registerScopeableFactory(getInstance());
      }
      else
      {
         throw new IllegalStateException("Initialize called > 1 time!");
      }
   }
   
   /**
    * Provides the singleton instance of this class as a
    * <code>Scopeable</code>.
    *
    * @return  This object.
    */
   public Scopeable createScopeable()
   {
      return this;
   }
   
   /**
    * Adds an accumulator to the current set of accumulators.
    * 
    * @param   accum
    *          The accumulator
    * @param   scope
    *          Block scope at which to register the accumulator:  CURRENT or
    *          NEXT are the only valid values.
    */
   public static void addAccumulator(Accumulator accum, ScopeLevel scope)
   {
      WorkArea wa = context.get();
      
      switch (scope)
      {
         case NEXT:
            if (wa.pendingSet == null)
            {
               wa.initPending();
            }
            wa.pendingSet.add(accum);
            break;
            
         case CURRENT:
            wa.stack.peek().add(accum);
            break;
      }
   }   
   
   /**
    * Provides a notification that a new scope is about to be entered.
    */
   public void scopeStart()
   {
      // ignore editing blocks
      if (TransactionManager.getBlockType() == BlockType.EDITING)
         return;
      
      WorkArea wa = context.get();
      Set<Accumulator> prevSet = wa.stack.peek();
      Set<Accumulator> currentSet = wa.pendingSet;
      
      // if no pending set exists, create one
      if (currentSet == null)
      {
         wa.initPending();
         currentSet = wa.pendingSet;
      }
      
      if (prevSet != null)
      {
         // roll up all accumulators from the previous scope to the new,
         // current scope
         currentSet.addAll(prevSet);
      }
      
      // make the pending set of accumulators the current scope's active set
      wa.stack.push(currentSet);
      
      // notify all registered accumulators a new scope is starting
      Iterator<Accumulator> itr = currentSet.iterator();
      while (itr.hasNext())
      {
         Accumulator a = itr.next();
         a.scopeStart();
      }
      
      // clear the pending set for the next scope
      wa.pendingSet = null;
   }
   
   /**
    * Provides a notification that a scope is about to be exited.
    */
   public void scopeFinished()
   {
      // ignore editing blocks
      if (TransactionManager.getBlockType() == BlockType.EDITING)
         return;

      WorkArea wa = context.get();
      
      Set<Accumulator> set = wa.stack.pop();
      if (set != null)
      {
         Iterator<Accumulator> itr = set.iterator();
         while (itr.hasNext())
         {
            Accumulator a = itr.next();
            a.scopeFinished();
         }
      }
   }

   /**
    * Provides notification that the external procedure scope has been deleted.
    * <p>
    * This is a no-op for accumulators.
    */
   @Override
   public void scopeDeleted()
   {
      // no-op
   }
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {
      /** Set of Accumulators to be added when the next block scope opens */
      private Set<Accumulator> pendingSet = null;
      
      /** Stack of Accumulator sets for each open block scope */
      private Deque<Set<Accumulator>> stack = 
         new ArrayDeque<Set<Accumulator>>();
      
      /**
       * Default constructor.
       */
      private WorkArea()
      {
         stack.push(new HashSet<Accumulator>(8));
      }
      
      /**
       * Initialize the set of pending Accumulator instances.
       */
      private void initPending()
      {
         pendingSet = new HashSet<Accumulator>(8);
      }
   }
   
   /**
    * Simple <code>ContextLocal</code> subclass that properly initializes the
    * context-local reference to a new instance of a case-sensitive
    * <code>ScopedSymbolDictionary</code> that has a global scope.
    */
   private static class ContextContainer
   extends ContextLocal<WorkArea>
   {      
      /**
       * Initializes the dictionary, the first time it is requested within a
       * new context.
       *
       * @return   The newly instantiated dictionary.
       */
      protected synchronized WorkArea initialValue()
      {
         WorkArea wa = new WorkArea();
         
         return wa;
      }   
   }   
}
